///////////////////////////////////////////////////////////////////////////////
//
//            (C) Copyright 2009-2010 OSR Open Systems Resources, Inc.
//            All Rights Reserved
//
//    This work is distributed under the OSR Non-Commercial Software License which is provided
//    at "http://www.osronline.com/page.cfm?name=NonCommLicense" in the hope that it will be
//    enlightening, but WITHOUT ANY WARRANTY; without even the implied warranty of MECHANTABILITY
//    or FITNESS FOR A PARTICULAR PURPOSE.
//
//            OSR Open Systems Resources, Inc.
//            105 Route 101A Suite 19
//            Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//
//
//
//    MODULE:
//
//        $File: //depot/tools/UMERESOURCES/UmEResourceDll/EResource.cpp $
//
//    ABSTRACT:
//
//     This module implements Executive Resources in user mode.
//
//    AUTHOR:
//
//        OSR Open Systems Resources, Inc.
// 
//    REVISION:   
//
//        $Revision: #1 $
//
//
///////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "UmEResourceDll.h"
#include "EResource.h"

#define OSR_LOCK_WAIT_IN_MILLISECONDS 200

//
// NTAPI's needed.  
//
extern "C" {

typedef enum _EVENT_TYPE {
	NotificationEvent,
	SynchronizationEvent
} EVENT_TYPE;

//
// We need to create a notification event, so that all threads waiting on the
// resource event will wakeup and reassesss their access to the resource.
//
NTSYSAPI
NTSTATUS
NTAPI
NtCreateEvent (
    OUT PHANDLE EventHandle,
    IN ACCESS_MASK DesiredAccess,
    IN POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL,
    IN EVENT_TYPE EventType,
    IN BOOLEAN InitialState
    );

typedef NTSTATUS (*PNTCREATEFILE)(
    OUT PHANDLE EventHandle,
    IN ACCESS_MASK DesiredAccess,
    IN POBJECT_ATTRIBUTES ObjectAttributes OPTIONAL,
    IN EVENT_TYPE EventType,
    IN BOOLEAN InitialState
    );

};


// OSREResource


///////////////////////////////////////////////////////////////////////////////
//
//  OSREResource
//
//      Constructor
//
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized before it can be used.
//
///////////////////////////////////////////////////////////////////////////////
OSREResource::OSREResource() 
: m_SharedCount(0)
, m_ExclusiveCount(0)
, m_SharedWaitersCount(0)
, m_ExclusiveWaitersCount(0)
, m_StarveExclusiveCount(0)
, m_WaitEventHandle(NULL)
, m_Initialized(FALSE)
{
}


///////////////////////////////////////////////////////////////////////////////
//
//  ~OSREResource
//
//      Destructor
//
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Deleted before the destructor is called.
//
///////////////////////////////////////////////////////////////////////////////
OSREResource::~OSREResource()
{
	ASSERT(!m_Initialized);
}


// OSREResource member functions


///////////////////////////////////////////////////////////////////////////////
//
//  AcquireResourceExclusive
//
//      Attempt to acquire the resource exclusively
//
//  INPUTS:
//		Wait - TRUE it wait to acquire, FALSE return immediately if cannnot
//				acquire the lock.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if acquired, FALSE otherwise.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN OSREResource::AcquireResourceExclusive(BOOLEAN Wait)
{
	DWORD cThread = GetCurrentThreadId();
	ULONG loopCount = 0;
	EnterCriticalSection(&m_CriticalSection);

	m_ExclusiveWaitersCount++;

	while(TRUE) {
		//
		//	See if we already own the resource.
		//
		CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

		if(ite != m_LockStateMap.end()) {
			//
			// We are in the list.   See what type of ownership we have.
			//
			if(ite->second.State & ACQUIRED_SHARED) {
				//
				// We have it shared.  This is bad, we can't acquire it.
				//
				RaiseException(0xC0000008,EXCEPTION_NONCONTINUABLE,0,NULL);
			} else if(ite->second.State & ACQUIRED_EXCLUSIVE) {
				//
				// We already own it exclusively, so just increment
				// our ownership count.
				//
				ite->second.AcquireCount++;
				m_ExclusiveCount++;
	            m_ExclusiveWaitersCount--;
				LeaveCriticalSection(&m_CriticalSection);
				return TRUE;
			}
		} else if(m_SharedCount || m_ExclusiveCount || m_StarveExclusiveCount) {  // ignore shared waiters
			//
			// The lock is either already owned by another thread shared, 
            // owned by another thread exclusively, or there is a sharedstarveExclusive
			// waiter.   Therefore we have to wait for the lock.   When the Event gets set
			// we will reaccess our access to the lock.
			//
            if(!Wait) {
                //
                // The caller does not want to wait, so get out of here.
                //
			    m_ExclusiveWaitersCount--;
			    LeaveCriticalSection(&m_CriticalSection);
                break;
            }
			LeaveCriticalSection(&m_CriticalSection);
			WaitForSingleObject(m_WaitEventHandle,INFINITE);
            //
            // Add a wait loop to prevent this thread from hogging
            // the CPU.   The event gets set when the lock holder
            // releases the lock and we have to ensure that
            // everyone gets a chance to run before continuing.
            //
			loopCount++;
			if(loopCount == 5) {
				SleepEx(OSR_LOCK_WAIT_IN_MILLISECONDS,TRUE);
				loopCount = 0;
			}
			EnterCriticalSection(&m_CriticalSection);
			continue;
		} else {
			//
			// Wow, we are free to acquire the lock.
			//
			OSRRESOURCE_STATE state;
			state.State = ACQUIRED_EXCLUSIVE;
			state.AcquireCount = 1;
			m_ExclusiveCount++;
			m_ExclusiveWaitersCount--;
			m_LockStateMap.insert(std::make_pair(cThread,state));
            ResetEvent(m_WaitEventHandle);
			LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		}
	}
	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  AcquireResourceShared
//
//      Attempt to acquire the resource shared
//
//  INPUTS:
//		Wait - TRUE it wait to acquire, FALSE return immediately if cannnot
//				acquire the lock.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if acquired, FALSE otherwise.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN OSREResource::AcquireResourceShared(BOOLEAN Wait)
{
	DWORD cThread = GetCurrentThreadId();
	ULONG loopCount = 0;
	EnterCriticalSection(&m_CriticalSection);

	m_SharedWaitersCount++;

	while(TRUE) {
		//
		// See if we already own the lock.
		//
		CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

		if(ite != m_LockStateMap.end()) {
			//
			// We already own it, let's see what access we already have.
			//
			ite->second.AcquireCount++;
			if(ite->second.State & ACQUIRED_SHARED) {
				//
				// We already have it shared, so just bump the count.
				//
				m_SharedCount++;
			} else if(ite->second.State & ACQUIRED_EXCLUSIVE) {
				//
				// We have it exclusive, so just give it to the
				// caller exclusive, 
				//
				m_ExclusiveCount++;
			}
			m_SharedWaitersCount--;
		    LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		} else if(m_ExclusiveCount || m_ExclusiveWaitersCount) {
			//
			// Someone else owns it exclusive or there are exclusive
			// waiters, so we have to wait.
			//
            if(!Wait) {
                //
                // The caller does not want to wait, so get out of here.
                //
			    m_SharedWaitersCount--;
			    LeaveCriticalSection(&m_CriticalSection);
                break;
            }
			LeaveCriticalSection(&m_CriticalSection);
			WaitForSingleObject(m_WaitEventHandle,INFINITE);
            //
            // Add a wait loop to prevent this thread from hogging
            // the CPU.   The event gets set when the lock holder
            // releases the lock and we have to ensure that
            // everyone gets a chance to run before continuing.
            //
			loopCount++;
			if(loopCount == 5) {
				SleepEx(OSR_LOCK_WAIT_IN_MILLISECONDS,TRUE);
				loopCount = 0;
			}
			EnterCriticalSection(&m_CriticalSection);
			continue;
		} else {
			//
			// Nobody has it or wants it, so we will take it.
			//
			OSRRESOURCE_STATE state;
			state.State = ACQUIRED_SHARED;
			state.AcquireCount = 1;
			m_SharedCount++;
			m_SharedWaitersCount--;
			m_LockStateMap.insert(std::make_pair(cThread,state));
            ResetEvent(m_WaitEventHandle);
			LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		}
	}
	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  AcquireSharedStarveExclusive
//
//      Attempt to acquire the resource shared, ignore Exclusive Waiters
//
//  INPUTS:
//		Wait - TRUE it wait to acquire, FALSE return immediately if cannnot
//				acquire the lock.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if acquired, FALSE otherwise.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN OSREResource::AcquireSharedStarveExclusive(BOOLEAN Wait)
{
	DWORD cThread = GetCurrentThreadId();
	ULONG loopCount = 0;
	EnterCriticalSection(&m_CriticalSection);

    //
    // Indicate that there is a thread that wants to acquire the lock
    // shared and start any exclusive waiters.
    //
    m_StarveExclusiveCount++;
	m_SharedWaitersCount++;

	while(TRUE) {
		//
		// See if we already own the lock.  We'd better.
		//
		CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

		if(ite != m_LockStateMap.end()) {
			//
			// We own it, let's check our access.
			//
			ite->second.AcquireCount++;
			if(ite->second.State & ACQUIRED_SHARED) {
				//
				// We already have it shared, so we'll just
				// bump the shared count.
				//
				m_SharedCount++;
			} else if(ite->second.State & ACQUIRED_EXCLUSIVE) {
				//
				// We already have it exclusive, so we'll just
				// give the caller exclusive access.
				//
				m_ExclusiveCount++;
			}
            m_StarveExclusiveCount--;
			LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		} else if(m_ExclusiveCount) {
			//
			//  Somebody owns it exclusive, so we have to wait.
			//
            if(!Wait) {
                //
                // The caller does not want to wait, so get out of here.
                //
                m_StarveExclusiveCount--;
	            m_SharedWaitersCount--;
			    LeaveCriticalSection(&m_CriticalSection);
                break;
            }
			LeaveCriticalSection(&m_CriticalSection);
			WaitForSingleObject(m_WaitEventHandle,INFINITE);
            //
            // Add a wait loop to prevent this thread from hogging
            // the CPU.   The event gets set when the lock holder
            // releases the lock and we have to ensure that
            // everyone gets a chance to run before continuing.
            //
			loopCount++;
			if(loopCount == 5) {
				SleepEx(OSR_LOCK_WAIT_IN_MILLISECONDS,TRUE);
				loopCount = 0;
			}
			EnterCriticalSection(&m_CriticalSection);
			continue;
		} else {
			//
			// No one owns the lock and we've ignored checking exclusive
			// waiters....
			//
			OSRRESOURCE_STATE state;
			state.State = ACQUIRED_SHARED;
			state.AcquireCount = 1;
			m_SharedCount++;
            m_StarveExclusiveCount--;
			m_SharedWaitersCount--;
			m_LockStateMap.insert(std::make_pair(cThread,state));
            ResetEvent(m_WaitEventHandle);
			LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		}
	}
	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  IsResourceAcquiredExclusive
//
//      See if the resource is acquired exclusively.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if acquired, FALSE otherwise.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN OSREResource::IsResourceAcquiredExclusive()
{
	DWORD cThread = GetCurrentThreadId();
	BOOLEAN	result = FALSE;

	EnterCriticalSection(&m_CriticalSection);

	//
	// See if we own the lock.
	//
	CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

	if(ite != m_LockStateMap.end()) {
		//
		// We own the lock, check our access.
		//
		if(ite->second.State & ACQUIRED_SHARED) {
			//
			// We own it shared, not exclusive.
			//
			LeaveCriticalSection(&m_CriticalSection);
			return FALSE;
		}
		//
		// We own it exclusive.
		//
		result = TRUE;
	}

	//
	// If we don't own the lock, result will be false by 
	// default, otherwise the if above will have set 
	// result.
	//
	LeaveCriticalSection(&m_CriticalSection);
	return result;
}


///////////////////////////////////////////////////////////////////////////////
//
//  IsResourceAcquiredShared
//
//      See if the resource is acquired shared.  If held exclusive, this is
//		also will return TRUE.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if acquired, FALSE otherwise.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN OSREResource::IsResourceAcquiredShared()
{
	DWORD cThread = GetCurrentThreadId();
	BOOLEAN	result = FALSE;

	EnterCriticalSection(&m_CriticalSection);

	//
	// See if we own the lock.
	//
	CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

	if(ite != m_LockStateMap.end()) {
		//
		// We own the lock, check our access.
		//
		if(ite->second.State & (ACQUIRED_SHARED | ACQUIRED_EXCLUSIVE)) {
			//
			// We own the lock, so return TRUE.
			//
			LeaveCriticalSection(&m_CriticalSection);
			return TRUE;
		}
		//
		// THis had better not happen.....
		//
		RaiseException(0xC0000001,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	//
	// If we get here, we don't own the lock and will return FALSE.
	//
	LeaveCriticalSection(&m_CriticalSection);
	return result;
}


///////////////////////////////////////////////////////////////////////////////
//
//  ReleaseResource
//
//      Release the resource.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned.
//
///////////////////////////////////////////////////////////////////////////////
VOID OSREResource::ReleaseResource()
{
	DWORD cThread = GetCurrentThreadId();

	EnterCriticalSection(&m_CriticalSection);

	//
	// See if we own the lock.  We'd better.....
	//
	CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

	if(ite != m_LockStateMap.end()) {
		//
		// We own it.   Decrement the ownership count based upon
		// our access.
		// 
		if(ite->second.State & ACQUIRED_SHARED) {
			m_SharedCount--;
		} else {
			m_ExclusiveCount--;
		}	

		//
		// Look at our ownership count in the lowpart of the
		// large integer.   If it is 1, then this is our
		// last reference to the lock, so we will delete
		// ourselves from the ownership map.
		//
		if(ite->second.AcquireCount == 1) {
			//
			// erase us, we no longer own the lock.
			//
			m_LockStateMap.erase(ite);

			//
			// Wake up anyone waiting on access to the lock.
			//
			SetEvent(m_WaitEventHandle);
		} else {
			//
			// We still have outstanding references....
			//
			ite->second.AcquireCount--;
		}
		LeaveCriticalSection(&m_CriticalSection);
	} else {
		RaiseException(0xC0000002,EXCEPTION_NONCONTINUABLE,0,NULL);
	}
}


///////////////////////////////////////////////////////////////////////////////
//
//  ConvertExclusiveToShared
//
//      Convert exclusive ownership of a resource to shared.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned.
//
///////////////////////////////////////////////////////////////////////////////
VOID OSREResource::ConvertExclusiveToShared()
{
	DWORD cThread = GetCurrentThreadId();

	EnterCriticalSection(&m_CriticalSection);

	//
	// See if we own the lock.  We'd better.
	//
	CMapThreadToLockState::iterator ite = m_LockStateMap.find(cThread);

	if(ite != m_LockStateMap.end()) {
		//
		// We own the lock, check our access.
		//
		if(ite->second.State != ACQUIRED_EXCLUSIVE) {
			//
			// We don't own it exclusive, this is bad.
			//
			RaiseException(0xC0000003,EXCEPTION_NONCONTINUABLE,0,NULL);
		} else {
			//
			// Make sure that we only have one reference on the lock
			// I don't know what to do if we one it more than once...
			//
//			if(ite->second.AcquireCount != 1) {
//				//
//				// We have it owned more than once.... Bad.
//				//
//				RaiseException(0xC0000004,EXCEPTION_NONCONTINUABLE,0,NULL);
//			} else {
				//
				// We own it once exclusively, convert it to
				// shared and update the counts.
				//
				ite->second.State = ACQUIRED_SHARED;
				m_SharedCount += ite->second.AcquireCount;
				m_ExclusiveCount -= ite->second.AcquireCount;

				//
				// Wake up anyone waiting for the lock.  If there are
				// other shared waiters they will be able to get the
				// lock now.
				//
				SetEvent(m_WaitEventHandle);
				LeaveCriticalSection(&m_CriticalSection);
				return;
//			}
		}
	}
	RaiseException(0xC0000005,EXCEPTION_NONCONTINUABLE,0,NULL);
}


///////////////////////////////////////////////////////////////////////////////
//
//  ReleaseResourceForThread
//
//      Release the resource owned by the input thread.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned.
//
///////////////////////////////////////////////////////////////////////////////
VOID OSREResource::ReleaseResourceForThread(ERESOURCE_THREAD ResourceThreadId)
{
	EnterCriticalSection(&m_CriticalSection);

	//
	// See if the input thread owns the lock.
	//
	CMapThreadToLockState::iterator ite = m_LockStateMap.find(ResourceThreadId);

	if(ite != m_LockStateMap.end()) {
		//
		// We found the input thread, update the counts based upon
		// the callers access.
		//
		if(ite->second.State & ACQUIRED_SHARED) {
			m_SharedCount--;
		} else {
			m_ExclusiveCount--;
		}				
		//
		// Look at the reference count in the lock.  If there is only one
		// reference left, then we can delete this owner from the map.
		//
		if(ite->second.AcquireCount == 1) {
			//
			// This is the callers last reference to the lock.  Erase
			// them and then signal all waiters to reaccess their lock
			// ownership.
			//
			m_LockStateMap.erase(ite);

			//
			// Wake up all waiters.
			//
			SetEvent(m_WaitEventHandle);
		} else {
			ite->second.AcquireCount--;
		}
		LeaveCriticalSection(&m_CriticalSection);
	} else {
		RaiseException(0xC0000006,EXCEPTION_NONCONTINUABLE,0,NULL);
	}
}


///////////////////////////////////////////////////////////////////////////////
//
//  InitializeResource
//
//      This routine initializes a resource for use.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		STATUS_SUCCESS if successful, an error otherwise.
//
//  NOTES:
//
//      The resource must be not be already Initialized.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OSREResource::InitializeResource()
{
	NTSTATUS status;

	if(!m_Initialized) {
#ifdef InitializeCriticalSectionAndSpinCount
		if(!InitializeCriticalSectionAndSpinCount(&m_CriticalSection,100)) {
			return STATUS_UNSUCCESSFUL;
		}
#else
		InitializeCriticalSection(&m_CriticalSection);
#endif
		//
		// Create a notification event.  All threads waiting on this event will be awaken
		// when this event is set.  These threads will then try to acquire the resource.
		//
#if 1
        HMODULE hNtDll = LoadLibrary(L"ntdll.dll");
        if(!hNtDll) {
			DeleteCriticalSection(&m_CriticalSection);
			return STATUS_UNSUCCESSFUL;
        }
        PNTCREATEFILE pNtCreateFile = (PNTCREATEFILE) GetProcAddress(hNtDll,"NtCreateEvent");
        if(!pNtCreateFile) {
			DeleteCriticalSection(&m_CriticalSection);
			return STATUS_UNSUCCESSFUL;
        }
		status = pNtCreateFile(&m_WaitEventHandle,GENERIC_ALL,NULL,NotificationEvent,FALSE);
#else
		status = NtCreateEvent(&m_WaitEventHandle,GENERIC_ALL,NULL,NotificationEvent,FALSE);
#endif
		if(status != STATUS_SUCCESS) {
			DeleteCriticalSection(&m_CriticalSection);
			return status;
		}

		if(!m_WaitEventHandle) {
			NTSTATUS error = GetLastError();
			DeleteCriticalSection(&m_CriticalSection);
			return STATUS_UNSUCCESSFUL;
		}

		m_Initialized = TRUE;
	}

	return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
//
//  DeleteResource
//
//      This routine deletes an initialized resource.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		STATUS_SUCCESS if successful, an error otherwise.
//
//  NOTES:
//
//      The resource must be Initialized, not owned, and not being waited for.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS OSREResource::DeleteResource()
{
	DWORD cThread = GetCurrentThreadId();

	EnterCriticalSection(&m_CriticalSection);

	if(m_SharedCount | m_ExclusiveCount |
		   m_SharedWaitersCount | m_ExclusiveWaitersCount |
		   !m_Initialized) {
		RaiseException(0xC0000007,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	DeleteCriticalSection(&m_CriticalSection);
	CloseHandle(m_WaitEventHandle);
	m_Initialized = FALSE;

	return STATUS_SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////
//
//  GetSharedWaiterCount
//
//      This routine returns the number of shared waiters.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		The number of shared waiters at the time of this call.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
ULONG	OSREResource::GetSharedWaiterCount()
{
	return m_SharedWaitersCount;
}


///////////////////////////////////////////////////////////////////////////////
//
//  GetExclusiveWaiterCount
//
//      This routine returns the number of exclusive waiters.
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		The number of exclusive waiters at the time of this call.
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
ULONG	OSREResource::GetExclusiveWaiterCount()
{
	return m_ExclusiveWaitersCount;
}


///////////////////////////////////////////////////////////////////////////////
//
//	C Interface routines.   This DLL exports a "C" interface to the resources.
//  All C++ constructs are hidden to the caller and are local to this DLL.
//
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//
//  OSRAcquireResourceExclusiveLite
//
//      This routine attempts to acquire a resource exclusively.
//
//  INPUTS:
//		Resource - address of resource.
//		Wait - TRUE if to wait for lock, FALSE if not
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if successful, FALSE otherwise
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
__cdecl OSRAcquireResourceExclusiveLite (
    __inout POSRRESOURCE Resource,
    __in BOOLEAN Wait
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->AcquireResourceExclusive(Wait);
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRAcquireResourceSharedLite
//
//      This routine attempts to acquire a resource shared.
//
//  INPUTS:
//		Resource - address of resource.
//		Wait - TRUE if to wait for lock, FALSE if not
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if successful, FALSE otherwise
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
__cdecl OSRAcquireResourceSharedLite (
    __inout POSRRESOURCE Resource,
    __in BOOLEAN Wait
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->AcquireResourceShared(Wait);
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRAcquireSharedStarveExclusive
//
//      This routine attempts to acquire a resource shared and starves any
//		exclusive waiters.
//
//  INPUTS:
//		Resource - address of resource.
//		Wait - TRUE if to wait for lock, FALSE if not
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if successful, FALSE otherwise
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
__cdecl OSRAcquireSharedStarveExclusive(
    __inout POSRRESOURCE Resource,
    __in BOOLEAN Wait
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->AcquireSharedStarveExclusive(Wait);
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRIsResourceAcquiredExclusiveLite
//
//      This routine checks to see if a resource is held exclusively by the
//		caller.
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if held, FALSE otherwise
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
__cdecl OSRIsResourceAcquiredExclusiveLite (
    __in POSRRESOURCE Resource
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->IsResourceAcquiredExclusive();
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRIsResourceAcquiredSharedLite
//
//      This routine checks to see if a resource is held shared by the
//		caller.  If held exclusive, TRUE will be returned also
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		TRUE if held, FALSE otherwise
//
//  NOTES:
//
//      The resource must be Initialized.
//
///////////////////////////////////////////////////////////////////////////////
BOOLEAN
__cdecl OSRIsResourceAcquiredSharedLite (
    __in POSRRESOURCE Resource
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->IsResourceAcquiredShared();
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRReleaseResourceLite
//
//      This routine releases a held lock.
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned.
//
///////////////////////////////////////////////////////////////////////////////
VOID
__cdecl OSRReleaseResourceLite(
    __inout POSRRESOURCE Resource
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	pResource->ReleaseResource();

	return;
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRConvertExclusiveToSharedLite
//
//      This routine converts exclusive ownership to shared ownership.
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned exclusively.
//
///////////////////////////////////////////////////////////////////////////////
VOID
__cdecl OSRConvertExclusiveToSharedLite(
    __inout POSRRESOURCE Resource
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	pResource->ConvertExclusiveToShared();

	return;
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRReleaseResourceForThreadLite
//
//      This routine releases the input threads ownership of a eresource.
//
//  INPUTS:
//		Resource - address of resource.
//		ResourceThreadId - thread Id.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		None.
//
//  NOTES:
//
//      The resource must be Initialized and owned.
//
///////////////////////////////////////////////////////////////////////////////
VOID
__cdecl OSRReleaseResourceForThreadLite(
    __inout POSRRESOURCE Resource,
    __in ERESOURCE_THREAD ResourceThreadId
    )
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	pResource->ReleaseResourceForThread(ResourceThreadId);

	return;
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRInitializeResourceLite
//
//      This routine initializes an eresource
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		STATUS_SUCCESS if initialized, an error otherwise.
//
//  NOTES:
//
//      The resource must be not be already Initialized.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS
__cdecl OSRInitializeResourceLite (
    __out POSRRESOURCE Resource
    )
{
	if(Resource->eResource) {
		RaiseException(0xC000000A,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pRes  = new OSREResource();

	Resource->eResource = (PVOID) pRes;

	if(!Resource->eResource) {
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	return pRes->InitializeResource();
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRDeleteResourceLite
//
//      This routine deletes an eresource
//
//  INPUTS:
//		Resource - address of resource.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		STATUS_SUCCESS if deleted, an error otherwise.
//
//  NOTES:
//
//      The resource must be already Initialized, not owned, and not waited for.
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS
__cdecl OSRDeleteResourceLite (
    __inout POSRRESOURCE Resource
    )
{
	NTSTATUS	status;

	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	status = pResource->DeleteResource();

	if(status == STATUS_SUCCESS) {
		delete pResource;
		Resource->eResource = NULL;
	}

	return status;
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRGetCurrentResourceThread
//
//      This routine returns the current thread
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		Current Thread Id.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
ERESOURCE_THREAD 
  __cdecl OSRGetCurrentResourceThread( 
    VOID
    )
{
	return GetCurrentThreadId();
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRGetSharedWaiterCount
//
//      This routine returns the number of shared waiters 
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		Returns the number of shared waiters.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
ULONG
__cdecl OSRGetSharedWaiterCount(
    __inout POSRRESOURCE Resource
	)
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->GetSharedWaiterCount();
}


///////////////////////////////////////////////////////////////////////////////
//
//  OSRGetSharedWaiterCount
//
//      This routine returns the number of shared waiters 
//
//  INPUTS:
//		None.
//
//  OUTPUTS:
//		None.
//
//  RETURNS:
//		Returns the number of shared waiters.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
ULONG
__cdecl OSRGetExclusiveWaiterCount(
    __inout POSRRESOURCE Resource
	)
{
	if(!Resource->eResource) {
		RaiseException(0xC0000009,EXCEPTION_NONCONTINUABLE,0,NULL);
	}

	OSREResource* pResource = (OSREResource*) Resource->eResource;

	return pResource->GetExclusiveWaiterCount();
}
